'use strict';

const spawnSync = require('child_process').spawnSync;

class ServerlessPlugin {
  constructor(serverless, options) {
    this.serverless = serverless;
    this.options = options;
    this.commands = {
      syncToS3: {
        usage: 'Deploys the `app` directory to your bucket',
        lifecycleEvents: [
          'sync',
        ],
      },
      invalidateCloudFront: {
        usage: 'Invalidates the cache of a cloudfront distribution',
        lifecycleEvents: [
          'invalidate',
        ],
      },
      bucketInfo: {
        usage: 'Fetches and prints out the deployed CloudFront bucket names',
        lifecycleEvents: [
          'bucketInfo',
        ],
      },
      domainInfo: {
        usage: 'Fetches and prints out the deployed CloudFront domain names',
        lifecycleEvents: [
          'domainInfo',
        ],
      },
    };

    this.hooks = {
      'syncToS3:sync': this.syncDirectory.bind(this),
      'invalidateCloudFront:invalidate': this.invalidateCloudFrontCache.bind(this),
      'domainInfo:domainInfo': this.domainInfo.bind(this),
      'bucketInfo:bucketInfo': this.bucketInfo.bind(this)
    };
  }

  getDescribeStacksOutput(outputKey) {
    const provider = this.serverless.getProvider('aws');
    const stackName = provider.naming.getStackName(this.options.stage);
    return provider
      .request(
        'CloudFormation',
        'describeStacks',
        { StackName: stackName },
        this.options.stage,
        this.options.region // eslint-disable-line comma-dangle
      )
      .then((result) => {
        const outputs = result.Stacks[0].Outputs;
        const output = outputs.find(entry => entry.OutputKey === outputKey);
        return output.OutputValue;
      });
  }

  invalidateCloudFrontCache(distribution) {
    this.getDescribeStacksOutput('WebAppCloudFrontDistributionId').then(distribution => {
      let args = [
        'cloudfront',
        'create-invalidation',
        '--distribution-id',
        distribution,
        '--paths',
        '/*'
      ]

      this.serverless.cli.log(args);
      const stresult = spawnSync('aws', args);
      const stout = stresult && stresult.stdout && stresult.stdout.toString();
      const sterr = stresult && stresult.stderr && stresult.stderr.toString();
      this.serverless.cli.log(stout || 'stdoud undefined');
      this.serverless.cli.log(sterr || 'stderr undefined');

      if (!sterr) {
        this.serverless.cli.log('Successfully invalidated cache');
      }
    });
  }

  constructCpArgs(s3Bucket, files, cache, maxAge=2592000) {
    let cacheControl = 'max-age=0,no-cache,no-store,must-revalidate';

    if (cache) {
      cacheControl = `max-age=${maxAge},public`;
    }

    const includes = files.reduce((acc, file) => {
      return [
        ...acc,
        '--include',
        file
      ];
    }, []);

    let cpArgs = [
      's3',
      'cp',
      `s3://${s3Bucket}/`,
      `s3://${s3Bucket}/`,
      '--exclude',
      '*',
      ...includes,
      '--recursive',
      '--metadata-directive',
      'REPLACE',
      '--cache-control',
      cacheControl,
      '--acl',
      'public-read'
    ]

    return cpArgs;
  }

  // syncs the `app` directory to the provided bucket
  syncDirectory() {
    this.getDescribeStacksOutput('WebAppS3BucketOutput').then(s3Bucket => {
      const profile = this.serverless.variables.service.provider.profile;
      const s3LocalPath = this.serverless.variables.service.custom.s3LocalPath;

      let rmArgs = [
        's3',
        'rm',
        `s3://${s3Bucket}`,
        '--recursive'
      ]

      this.serverless.cli.log(rmArgs);
      const rmResult = spawnSync('aws', rmArgs);
      const rmStout = rmResult && rmResult.stdout && rmResult.stdout.toString();
      const rmSterr = rmResult && rmResult.stderr && rmResult.stderr.toString();
      this.serverless.cli.log(rmStout || 'stdoud undefined');
      this.serverless.cli.log(rmSterr || 'stderr undefined');

      let sterr = null
      if (!rmSterr) {
        let args = [
          's3',
          'sync',
          s3LocalPath,
          `s3://${s3Bucket}/`,
        ];


        if (profile) {
          args = args.concat([
            '--profile',
            profile
          ]);
        }

        this.serverless.cli.log(args);
        const result = spawnSync('aws', args);
        const stdout = result && result.stdout && result.stdout.toString();
        sterr = result && result.stderr && result.stderr.toString();
        this.serverless.cli.log(stdout || 'stdoud undefined');
        this.serverless.cli.log(sterr || 'stderr undefined');
      }

      let cpSterr = null;

      const cacheSettings = [
        {
          files: ['index.html', 'service-worker.js'],
          cache: false
        },
        {
          files: ['*.js', '*.css'],
          cache: true
        }
      ];

      if (!sterr && !cpSterr) {
        cacheSettings.forEach(({ files, cache, maxAge=2592000 }) => {
          let cpArgs = this.constructCpArgs(s3Bucket, files, cache, maxAge);

          if (profile) {
            cpArgs = cpArgs.concat([
              '--profile',
              profile
            ]);
          }

          this.serverless.cli.log(cpArgs);
          const cpResult = spawnSync('aws', cpArgs);
          const cpStdout = cpResult && cpResult.stdout && cpResult.stdout.toString();
          const cpSterr = cpResult && cpResult.stderr && cpResult.stderr.toString();
          this.serverless.cli.log(cpStdout || 'stdoud undefined');
          this.serverless.cli.log(cpSterr || 'stderr undefined');
        });
      }

      if (!cpSterr) {
        this.serverless.cli.log('Successfully synced to the S3 bucket');
      }
    });
  }

  // fetches the bucket name from the CloudFront outputs and prints it out
  bucketInfo() {
    this.getDescribeStacksOutput('WebAppS3BucketOutput').then(outputValue =>
      this.serverless.cli.log(`Web App Bucket: ${outputValue || 'Not Found'}`)
    );
  }

  // fetches the domain name from the CloudFront outputs and prints it out
  domainInfo() {
    this.getDescribeStacksOutput('WebAppCloudFrontDistributionOutput').then(outputValue =>
      this.serverless.cli.log(`Web App Domain: ${outputValue || 'Not Found'}`)
    );
  }
}

module.exports = ServerlessPlugin;
